pub mod watcher;

use std::{
	error::Error,
	collections::*,
	convert::Infallible,
	iter::FromIterator,
	net::SocketAddr,
	sync::{
		Arc
	},
};

use dashmap::DashMap;
use futures::channel::oneshot;
use handlebars::Handlebars;
use hyper::{
	Body, 
	Method,
	Request, 
	Response, 
	Server, 
	StatusCode,
};
use hyper::service::{make_service_fn, service_fn};
use serde::{
	Deserialize, 
	Serialize,
};
use tokio::{
	sync::Mutex,
};

use crate::{
	http_serde,
	prefix_match,
};

/*

Here's what we need to handle:

When a request comes in:

- Park the client in response_rendezvous
- Look up the server ID in request_rendezvous
- If a server is parked, unpark it and send the request
- Otherwise, queue the request

When a server comes to listen:

- Look up the server ID in request_rendezvous
- Either return all pending requests, or park the server

When a server comes to respond:

- Look up the parked client in response_rendezvous
- Unpark the client and begin streaming

So we need these lookups to be fast:

- Server IDs, where (1 server) or (0 or many clients)
can be parked
- Request IDs, where 1 client is parked

*/

enum RequestRendezvous {
	ParkedClients (Vec <http_serde::WrappedRequest>),
	ParkedServer (oneshot::Sender <http_serde::WrappedRequest>),
}

type ResponseRendezvous =  oneshot::Sender <(http_serde::ResponseParts, Body)>;

// Stuff we need to load from the config file and use to
// set up the HTTP server

#[derive (Default, Deserialize)]
pub struct ConfigFile {
	pub port: Option <u16>,
	pub server_tripcodes: HashMap <String, String>,
}

// Stuff we actually need at runtime

struct Config {
	server_tripcodes: HashMap <String, blake3::Hash>,
}

impl From <&ConfigFile> for Config {
	fn from (f: &ConfigFile) -> Self {
		let trips = HashMap::from_iter (f.server_tripcodes.iter ()
		.map (|(k, v)| {
			use std::convert::TryInto;
			let bytes: Vec <u8> = base64::decode (v).unwrap ();
			let bytes: [u8; 32] = (&bytes [..]).try_into ().unwrap ();
			
			let v = blake3::Hash::from (bytes);
			
			(k.clone (), v)
		}));
		
		Self {
			server_tripcodes: trips,
		}
	}
}

pub struct RelayState {
	config: Config,
	handlebars: Arc <Handlebars <'static>>,
	
	// Key: Server ID
	request_rendezvous: Mutex <HashMap <String, RequestRendezvous>>,
	
	// Key: Request ID
	response_rendezvous: DashMap <String, ResponseRendezvous>,
}

impl Default for RelayState {
	fn default () -> Self {
		Self {
			config: Config::from (&ConfigFile::default ()),
			handlebars: Arc::new (load_templates ().unwrap ()),
			request_rendezvous: Default::default (),
			response_rendezvous: Default::default (),
		}
	}
}

impl From <&ConfigFile> for RelayState {
	fn from (config_file: &ConfigFile) -> Self {
		Self {
			config: Config::from (config_file),
			handlebars: Arc::new (load_templates ().unwrap ()),
			request_rendezvous: Default::default (),
			response_rendezvous: Default::default (),
		}
	}
}

impl RelayState {
	pub async fn list_servers (&self) -> Vec <String> {
		self.request_rendezvous.lock ().await.iter ()
		.map (|(k, _)| (*k).clone ())
		.collect ()
	}
}

fn status_reply <B: Into <Body>> (status: StatusCode, b: B)
-> Response <Body>
{
	Response::builder ().status (status).body (b.into ()).unwrap ()
}

async fn handle_http_listen (
	state: Arc <RelayState>, 
	watcher_code: String,
	api_key: &[u8],
)
-> Response <Body>
{
	let trip_error = status_reply (StatusCode::UNAUTHORIZED, "Bad X-ApiKey");
	
	let expected_tripcode = match state.config.server_tripcodes.get (&watcher_code) {
		None => {
			eprintln! ("Denied http_listen for non-existent server name {}", watcher_code);
			return trip_error;
		},
		Some (x) => x,
	};
	let actual_tripcode = blake3::hash (api_key);
	
	if expected_tripcode != &actual_tripcode {
		eprintln! ("Denied http_listen for bad tripcode {}", base64::encode (actual_tripcode.as_bytes ()));
		return trip_error;
	}
	
	use RequestRendezvous::*;
	
	let (tx, rx) = oneshot::channel ();
	
	{
		let mut request_rendezvous = state.request_rendezvous.lock ().await;
		
		if let Some (ParkedClients (v)) = request_rendezvous.remove (&watcher_code) 
		{
			return status_reply (StatusCode::OK, rmp_serde::to_vec (&v).unwrap ());
		}
		
		request_rendezvous.insert (watcher_code, ParkedServer (tx));
	}
	
	let one_req = vec! [
		rx.await.unwrap (),
	];
	
	return status_reply (StatusCode::OK, rmp_serde::to_vec (&one_req).unwrap ());
}

async fn handle_http_response (
	req: Request <Body>,
	state: Arc <RelayState>, 
	req_id: String,
)
	-> Response <Body>
{
	let (parts, body) = req.into_parts ();
	let resp_parts: http_serde::ResponseParts = rmp_serde::from_read_ref (&base64::decode (parts.headers.get (crate::PTTH_MAGIC_HEADER).unwrap ()).unwrap ()).unwrap ();
	
	match state.response_rendezvous.remove (&req_id) {
		Some ((_, tx)) => {
			match tx.send ((resp_parts, body)) {
				Ok (()) => status_reply (StatusCode::OK, "Connected to remote client...\n"),
				_ => status_reply (StatusCode::BAD_GATEWAY, "Failed to connect to client"),
			}
			
		},
		None => status_reply (StatusCode::BAD_REQUEST, "Request ID not found in response_rendezvous"),
	}
}

async fn handle_http_request (
	req: http::request::Parts,
	uri: String,
	state: Arc <RelayState>, 
	watcher_code: String
)
	-> Response <Body>
{
	if ! state.config.server_tripcodes.contains_key (&watcher_code) {
		return status_reply (StatusCode::NOT_FOUND, "Unknown server");
	}
	
	let req = match http_serde::RequestParts::from_hyper (req.method, uri, req.headers) {
		Ok (x) => x,
		_ => return status_reply (StatusCode::BAD_REQUEST, "Bad request"),
	};
	
	let (tx, rx) = oneshot::channel ();
	
	let id = ulid::Ulid::new ().to_string ();
	state.response_rendezvous.insert (id.clone (), tx);
	
	{
		let mut request_rendezvous = state.request_rendezvous.lock ().await;
		
		let wrapped = http_serde::WrappedRequest {
			id,
			req,
		};
		
		use RequestRendezvous::*;
		
		let new_rendezvous = match request_rendezvous.remove (&watcher_code) {
			Some (ParkedClients (mut v)) => {
				v.push (wrapped);
				ParkedClients (v)
			},
			Some (ParkedServer (s)) => {
				// If sending to the server fails, queue it
				
				match s.send (wrapped) {
					Ok (()) => ParkedClients (vec! []),
					Err (wrapped) => ParkedClients (vec! [wrapped]),
				}
			},
			None => ParkedClients (vec! [wrapped]),
		};
		
		request_rendezvous.insert (watcher_code, new_rendezvous);
	}
	
	let timeout = tokio::time::delay_for (std::time::Duration::from_secs (30));
	
	let received = tokio::select! {
		val = rx => val,
		() = timeout => {
			return status_reply (StatusCode::GATEWAY_TIMEOUT, "Remote server never responded")
		},
    };
	
	match received {
		Ok ((parts, body)) => {
			let mut resp = Response::builder ()
			.status (hyper::StatusCode::from (parts.status_code));
			
			for (k, v) in parts.headers.into_iter () {
				resp = resp.header (&k, v);
			}
			
			resp.body (body)
			.unwrap ()
		},
		_ => status_reply (StatusCode::GATEWAY_TIMEOUT, "Remote server timed out"),
	}
}

async fn handle_all (req: Request <Body>, state: Arc <RelayState>) 
-> Result <Response <Body>, Infallible>
{
	let path = req.uri ().path ();
	//println! ("{}", path);
	
	let api_key = req.headers ().get ("X-ApiKey");
	
	if req.method () == Method::POST {
		// This is stuff the server can use. Clients can't
		// POST right now
		
		return Ok (if let Some (request_code) = prefix_match (path, "/7ZSFUKGV_http_response/") {
			let request_code = request_code.into ();
			handle_http_response (req, state, request_code).await
		}
		else {
			status_reply (StatusCode::BAD_REQUEST, "Can't POST this\n")
		});
	}
	
	Ok (if let Some (listen_code) = prefix_match (path, "/7ZSFUKGV_http_listen/") {
		let api_key = match api_key {
			None => return Ok (status_reply (StatusCode::UNAUTHORIZED, "Can't register as server without an API key")),
			Some (x) => x,
		};
		handle_http_listen (state, listen_code.into (), api_key.as_bytes ()).await
	}
	else if let Some (rest) = prefix_match (path, "/servers/") {
		if rest == "" {
			#[derive (Serialize)]
			struct ServerEntry <'a> {
				path: &'a str,
				name: &'a str,
			}
			
			#[derive (Serialize)]
			struct ServerListPage <'a> {
				servers: Vec <ServerEntry <'a>>,
			}
			
			let names = state.list_servers ().await;
			
			//println! ("Found {} servers", names.len ());
			
			let page = ServerListPage {
				servers: names.iter ()
				.map (|name| ServerEntry {
					name: &name,
					path: &name,
				})
				.collect (),
			};
			
			let s = state.handlebars.render ("relay_server_list", &page).unwrap ();
			status_reply (StatusCode::OK, s)
		}
		else if let Some (idx) = rest.find ('/') {
			let listen_code = String::from (&rest [0..idx]);
			let path = String::from (&rest [idx..]);
			let (parts, _) = req.into_parts ();
			
			handle_http_request (parts, path, state, listen_code).await
		}
		else {
			status_reply (StatusCode::BAD_REQUEST, "Bad URI format")
		}
	}
	else if path == "/relay_up_check" {
		status_reply (StatusCode::OK, "Relay is up\n")
	}
	else {
		status_reply (StatusCode::OK, "Hi\n")
	})
}

pub fn load_templates () 
-> Result <Handlebars <'static>, Box <dyn Error>>
{
	let mut handlebars = Handlebars::new ();
	handlebars.set_strict_mode (true);
	
	for (k, v) in vec! [
		("relay_server_list", "relay_server_list.html"),
	].into_iter () {
		handlebars.register_template_file (k, format! ("ptth_handlebars/{}", v))?;
	}
	
	Ok (handlebars)
}

pub async fn run_relay (
	state: Arc <RelayState>
) 
	-> Result <(), Box <dyn Error>> 
{
	let addr = SocketAddr::from ((
		[0, 0, 0, 0],
		4000,
	));
	
	{
		let mut tripcode_set = HashSet::new ();
		
		for (_, v) in state.config.server_tripcodes.iter () {
			if ! tripcode_set.insert (v) {
				panic! ("Two servers have the same tripcode. That is not allowed.");
			}
		}
	}
	
	eprintln! ("Loaded {} server tripcodes", state.config.server_tripcodes.len ());
	
	let make_svc = make_service_fn (|_conn| {
		let state = state.clone ();
		
		async {
			Ok::<_, Infallible> (service_fn (move |req| {
				let state = state.clone ();
				
				handle_all (req, state)
			}))
		}
	});
	
	let server = Server::bind (&addr).serve (make_svc);
	
	server.await?;
	
	Ok (())
}

pub async fn main (config_file: ConfigFile) 
-> Result <(), Box <dyn Error>> 
{
	let state = Arc::new (RelayState::from (&config_file));
	
	run_relay (state).await
}

#[cfg (test)]
mod tests {
	
}
